/* --- BEGIN COPYRIGHT BLOCK ---
 * Copyright (C) 2016  Red Hat
 * see files 'COPYING' and 'COPYING.openssl' for use and warranty
 * information
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Additional permission under GPLv3 section 7:
 * 
 * If you modify this Program, or any covered work, by linking or
 * combining it with OpenSSL, or a modified version of OpenSSL licensed
 * under the OpenSSL license
 * (https://www.openssl.org/source/license.html), the licensors of this
 * Program grant you additional permission to convey the resulting
 * work. Corresponding Source for a non-source form of such a
 * combination shall include the source code for the parts that are
 * licensed under the OpenSSL license as well as that of the covered
 * work.
 * --- END COPYRIGHT BLOCK ---
 */

/*
 * A self hosting echo server, that stress tests job addition,
 * removal, timers, and more.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/* For cmocka */
#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>

#include <nspr.h>
#include <plstr.h>
#include <prlog.h>

#include "nunc-stans.h"

#include <stdio.h>
#include <signal.h>

#include <syslog.h>
#include <string.h>
#include <inttypes.h>

#include <time.h>
#include <sys/time.h>

struct conn_ctx {
    size_t offset; /* current offset into buffer for reading or writing */
    size_t len; /* size of buffer */
    size_t needed; /* content-length + start of body */
    size_t body; /* when reading, offset from buffer of beginning of http body */
    size_t cl; /* http content-length when reading */
#define CONN_BUFFER_SIZE BUFSIZ /* default buffer size */
    char *buffer;
    PRBool need_sec_layer_before_next_read;
    struct ns_sec_ctx_t *sc;
};

static FILE *logfp;

void do_logging(int, const char*, ...);

PRInt32 client_success_count = 0;
PRInt32 server_success_count = 0;
PRInt32 client_fail_count = 0;
PRInt32 client_timeout_count = 0;
PRInt32 server_fail_count = 0;
PRInt32 job_count = 0;
PRInt32 client_thread_count = 80;
PRInt32 server_thread_count = 20;
PRInt32 jobs = 200;
PRInt32 test_timeout = 70;

#define PR_WOULD_BLOCK(iii) (iii == PR_PENDING_INTERRUPT_ERROR) || (iii == PR_WOULD_BLOCK_ERROR)

static void
setup_logging( void )
{
    logfp = stdout;
}

static void
do_vlogging(int level, const char *format, va_list varg)
{
#ifdef DEBUG
    if (level <= LOG_DEBUG) {
#else
    if (level <= LOG_ERR) {
#endif
        fprintf(logfp, "%d ", level);
        vfprintf(logfp, format, varg);
    }
}

void
do_logging(int level, const char *format, ...)
{
    va_list varg;
    va_start(varg, format);
    do_vlogging(level, format, varg);
    va_end(varg);
}

/* Server specifics */

static struct conn_ctx *
conn_ctx_new( void )
{
    struct conn_ctx *connctx = PR_NEW(struct conn_ctx);
    connctx->offset = 0;
    connctx->len = 0;
    connctx->buffer = NULL;
    connctx->needed = 0;
    connctx->body = 0;
    connctx->cl = 0;
    connctx->need_sec_layer_before_next_read = PR_FALSE;
    connctx->sc = NULL;
    return connctx;
}

static void
conn_ctx_free(struct conn_ctx *connctx)
{
    /* Why don't we use PR_DELETE here? */
    if (connctx->buffer != NULL) {
        PR_Free(connctx->buffer);
        connctx->buffer = NULL;
    }
    PR_DELETE(connctx);

}

static void
server_conn_write(struct ns_job_t *job)
{
    struct conn_ctx *connctx;
    PRInt32 len;

    do_logging(LOG_DEBUG, "job about to write ...\n");
    PR_ASSERT(job);
    PR_ASSERT(ns_job_get_data(job));
    connctx = (struct conn_ctx *)ns_job_get_data(job);
    if (NS_JOB_IS_TIMER(ns_job_get_output_type(job))) {
        do_logging(LOG_ERR, "conn_write: job [%p] timeout\n", job);
        conn_ctx_free(connctx);
        ns_job_done(job);
        return;
    }

    /* Get the data out of our connctx */
    char *data = PR_Malloc(sizeof(char) * (connctx->offset + 1));
    memcpy(data, connctx->buffer, connctx->offset);
    data[connctx->offset] = '\0';

    /* Should I write a new line also */
    len = PR_Write(ns_job_get_fd(job), data, connctx->offset);

    /* Set the buffer window back to the start */
    connctx->offset = 0;

    if (len < 0) {
        /* Error */
        printf("ERROR: occured in conn_write\n");
    }
    /* After we finish writing, do we stop being a thread? */
    do_logging(LOG_DEBUG, "Wrote \"%s\"\n", data);
    free(data);
    /* The job is still a *read* IO event job, so this should be okay */
    ns_job_rearm(job);
    return;
}

static void
server_conn_read(struct ns_job_t *job)
{
    do_logging(LOG_DEBUG, "Reading from connection\n");

    struct conn_ctx *connctx;
    PRInt32 len, nbytes;

    PR_ASSERT(job);
    PR_ASSERT(ns_job_get_data(job));

    connctx = (struct conn_ctx *)ns_job_get_data(job);
    if (NS_JOB_IS_TIMER(ns_job_get_output_type(job))) {
        /* The event that triggered this call back is because we timed out waiting for IO */
        do_logging(LOG_ERR, "conn_read: job [%p] timed out\n", job);
        conn_ctx_free(connctx);
        ns_job_done(job);
        return;
    }

    if (connctx->needed != 0) {
        nbytes = connctx->needed - connctx->offset;
    } else {
        nbytes = CONN_BUFFER_SIZE;
    }

    /* If our buffer is incorrectly sized, realloc it to match what we are about to read */
    if ((nbytes + connctx->offset) > connctx->len) {
        connctx->len = nbytes + connctx->offset;
        connctx->buffer = (char *)PR_Realloc(connctx->buffer, connctx->len * sizeof(char));
    }

    /* Read and append to our buffer */
    len = PR_Read(ns_job_get_fd(job), connctx->buffer + connctx->offset, nbytes);
    if (len < 0) {
        PRErrorCode prerr = PR_GetError();
        if (PR_WOULD_BLOCK(prerr)) {
            /* We don't have poll, so we rearm the job for more data */
            if (NS_JOB_IS_PERSIST(ns_job_get_type(job)) != 0) {
                ns_job_rearm(job);
            }
            do_logging(LOG_ERR, "conn_read: block error for job [%p] %d: %s\n", job, PR_GetError(), PR_ErrorToString(PR_GetError(), PR_LANGUAGE_I_DEFAULT));
            return;
        } else {
            do_logging(LOG_ERR, "conn_read: read error for job [%p] %d: %s\n", job, PR_GetError(), PR_ErrorToString(PR_GetError(), PR_LANGUAGE_I_DEFAULT));
            conn_ctx_free(connctx);
            ns_job_done(job);
            return;
        }
        /* Error */
    } else if (len == 0) {
        /* Didn't read anything */
        do_logging(LOG_DEBUG, "conn_read: job [%p] closed\n", job);
        /* Increment the success */
        PR_AtomicAdd(&server_success_count, 1);
        conn_ctx_free(connctx);
        ns_job_done(job);
        return;
    } else {
        /* Wild data appears! */
        connctx->offset += len;
        char *data = PR_Malloc(sizeof(char) * (connctx->offset + 1));
        memcpy(data, connctx->buffer, connctx->offset);
        data[connctx->offset] = '\0';
        do_logging(LOG_DEBUG, "[%p] data received \n", job);
        do_logging(LOG_DEBUG, "data: \"%s\" \n", data);
        free(data);
        server_conn_write(job);
        do_logging(LOG_DEBUG, "job rearmed for write ...\n");
        return;
    }
}

static void
server_conn_handler(struct ns_job_t *job)
{
    do_logging(LOG_DEBUG, "Handling a connection\n");

    PR_ASSERT(job);

    if (NS_JOB_IS_READ(ns_job_get_type(job)) != 0) {
        server_conn_read(job);
    } else {
        /* We should not be able to get here! */
    }

    return;
}

static void
server_listen_accept(struct ns_job_t *job)
{
    PRFileDesc *connfd = NULL;
    struct conn_ctx *connctx;
    PRSocketOptionData prsod = {PR_SockOpt_Nonblocking, {PR_TRUE}};

    PR_ASSERT(job);

    connfd = PR_Accept(ns_job_get_fd(job), NULL, PR_INTERVAL_NO_WAIT);

    if (connfd != NULL) {
        PR_SetSocketOption(connfd, &prsod);
        connctx = conn_ctx_new();

        ns_add_io_job(ns_job_get_tp(job), connfd, NS_JOB_READ|NS_JOB_THREAD, server_conn_handler, connctx, NULL);

        do_logging(LOG_DEBUG, "server_listen_accept: accepting connection to job [%p]\n", job);

    } else {
        PRErrorCode prerr = PR_GetError();
        if (PR_WOULD_BLOCK(prerr)) {
            /* Let it go .... let it gooooo! */
            /* Can't hold up connection dispatch anymore! */
        } else {
            do_logging(LOG_ERR, "server_listen_accept: accept error for job [%p] %d %s\n", job, prerr, PR_ErrorToString(prerr, PR_LANGUAGE_I_DEFAULT));
        }
    }
}

static struct ns_job_t *
server_listener_init(ns_thrpool_t *tp, PRFileDesc *listenfd)
{
    struct ns_job_t *listen_job = NULL;
    ns_add_io_job(tp, listenfd, NS_JOB_ACCEPT|NS_JOB_PERSIST, server_listen_accept, NULL, &listen_job);
    return listen_job;
}

/* Client specifics */

static void
test_client_shutdown(struct ns_job_t *job)
{
    do_logging(LOG_DEBUG, "Received shutdown signal\n");
    do_logging(LOG_DEBUG, "status .... job_count: %d fail_count: %d success_count: %d\n", job_count, client_fail_count, client_success_count);
    ns_thrpool_shutdown(ns_job_get_tp(job));
    /* This also needs to start the thrpool shutdown for the server. */
    ns_thrpool_shutdown(ns_job_get_data(job));

}

static void
client_response_cb(struct ns_job_t *job)
{

    char *buffer = PR_Malloc(20);
    PRInt32 buflen = 20;
    PRInt32 len = 0;

    len = PR_Read(ns_job_get_fd(job), buffer, buflen);
    if (len < 0) {
        /* PRErrorCode prerr = PR_GetError(); */
        do_logging(LOG_ERR, "FAIL: connection error, no data \n");
        PR_AtomicAdd(&client_fail_count, 1);
        goto done;
    } else if (len == 0) {
        do_logging(LOG_ERR, "FAIL: connection closed, no data \n");
        PR_AtomicAdd(&client_fail_count, 1);
        goto done;
    } else {
        /* Be paranoid, force last byte null */
        buffer[buflen - 1] = '\0';
        if (strncmp("this is a test!\n", buffer, strlen("this is a test!\n")) != 0)
        {
            do_logging(LOG_ERR, "FAIL: connection incorrect response, no data \n");
            PR_AtomicAdd(&client_fail_count, 1);
            goto done;
        }
    }

    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);

    PR_AtomicAdd(&client_success_count, 1);
    do_logging(LOG_ERR, "PASS: %ld.%ld %d\n", ts.tv_sec, ts.tv_nsec, client_success_count);

done:
    PR_Free(buffer);
    ns_job_done(job);
}

static void
client_initiate_connection_cb(struct ns_job_t *job)
{
    /* Create a socket */
    PRFileDesc *sock = NULL;
    PRNetAddr netaddr;
    char *data = "this is a test!\n";

    sock = PR_OpenTCPSocket(PR_AF_INET6);
    if (sock == NULL) {
        do_logging(LOG_ERR, "Socket failed\n");
        PR_AtomicAdd(&client_fail_count, 1);
        goto done;
    }

    memset(&netaddr, 0, sizeof(netaddr));
    PR_SetNetAddr(PR_IpAddrLoopback, PR_AF_INET6, 12345, &netaddr);

    /* Connect */
    /*  */
    if (PR_Connect(sock, &netaddr, PR_SecondsToInterval(2)) != 0) {
        /* char *err = malloc(PR_GetErrorTextLength()); */
        char *err = NULL;
        PR_GetErrorText(err);
        do_logging(LOG_ERR, "FAIL: cannot connect, timeout %d -> %s \n", PR_GetError(), err);
        /* Atomic increment fail */
        PR_AtomicAdd(&client_timeout_count, 1);

        if (sock != NULL) {
            PR_Close(sock);
        }
        goto done;
    }
    /* Now write data. */
    PR_Write(sock, data, strlen(data) + 1);
    /* create the read job to respond to events on the socket. */
    ns_add_io_job(ns_job_get_tp(job), sock, NS_JOB_READ|NS_JOB_THREAD, client_response_cb, NULL, NULL);

done:
    ns_job_done(job);
}

static void
client_create_work(struct ns_job_t *job)
{
    struct timespec ts;
    PR_Sleep(PR_SecondsToInterval(1));
    clock_gettime(CLOCK_MONOTONIC, &ts);
    printf("BEGIN: %ld.%ld\n", ts.tv_sec, ts.tv_nsec);
    for (PRUint64 i = 0; i < jobs; i++) {
        ns_add_job(ns_job_get_tp(job), NS_JOB_NONE|NS_JOB_THREAD, client_initiate_connection_cb, NULL, NULL);
    }
    ns_job_done(job);

    printf("Create work thread complete!\n");
}

static void
ns_stress_test(void **state)
{

    /* Setup both thread pools. */

    /* Client first */

    PRInt32 job_count = jobs * client_thread_count;
    struct ns_thrpool_t *ctp;
    struct ns_thrpool_config client_ns_config;
    struct ns_job_t *sigterm_job = NULL;
    struct ns_job_t *sighup_job = NULL;
    struct ns_job_t *sigint_job = NULL;
    struct ns_job_t *final_job = NULL;

    struct timeval timeout = { test_timeout, 0 };

    setup_logging();

    ns_thrpool_config_init(&client_ns_config);
    client_ns_config.max_threads = client_thread_count;
    client_ns_config.log_fct = do_vlogging;
    ctp = ns_thrpool_new(&client_ns_config);

    /* Now the server */

    struct ns_thrpool_t *stp;
    struct ns_thrpool_config server_ns_config;
    struct ns_job_t *listen_job = NULL;

    ns_thrpool_config_init(&server_ns_config);
    server_ns_config.max_threads = server_thread_count;
    server_ns_config.log_fct = do_vlogging;
    stp = ns_thrpool_new(&server_ns_config);

    /* Now, add the signal handlers. */

    ns_add_signal_job(ctp, SIGTERM, NS_JOB_THREAD, test_client_shutdown, stp, &sigterm_job);
    ns_add_signal_job(ctp, SIGHUP, NS_JOB_THREAD, test_client_shutdown, stp, &sighup_job);
    ns_add_signal_job(ctp, SIGINT, NS_JOB_THREAD, test_client_shutdown, stp, &sigint_job);

    /* Create the socket for the server, and set it to listen. */

    PRFileDesc *listenfd = NULL;
    PRNetAddr netaddr;
    PRSocketOptionData prsod = {PR_SockOpt_Nonblocking, {PR_TRUE}};

    PR_SetNetAddr(PR_IpAddrAny, PR_AF_INET6, 12345, &netaddr);

    listenfd = PR_OpenTCPSocket(PR_NetAddrFamily(&netaddr));
    PR_SetSocketOption(listenfd, &prsod);
    prsod.option = PR_SockOpt_Reuseaddr;
    PR_SetSocketOption(listenfd, &prsod);
    PR_Bind(listenfd, &netaddr);
    PR_Listen(listenfd, 32);

    listen_job = server_listener_init(stp, listenfd);

    /* Add the timeout. */

    ns_add_timeout_job(ctp, &timeout, NS_JOB_NONE|NS_JOB_THREAD, test_client_shutdown, stp, &final_job);

    /* While true, add connect / write jobs */
    for (PRUint64 i = 0; i < client_thread_count; i++) {
        ns_add_job(ctp, NS_JOB_NONE|NS_JOB_THREAD, client_create_work, NULL, NULL);
    }

    /* Wait for all the clients to be done dispatching jobs to the server */

    if (ns_thrpool_wait(ctp) != 0) {
        printf("Error in ctp?\n");
    }

    if (ns_thrpool_wait(stp) != 0) {
        printf("Error in stp?\n");
    }

    /* Is this valid, because this is persistent? */
    ns_job_done(listen_job);


    ns_job_done(sigterm_job);
    ns_job_done(sighup_job);
    ns_job_done(sigint_job);
    ns_job_done(final_job);

    ns_thrpool_destroy(stp);
    ns_thrpool_destroy(ctp);

    /* Destroy the server thread pool. */

    if (client_success_count != job_count) {
        do_logging(LOG_ERR, "FAIL, not all client jobs succeeded!\n");
    }
    if (server_success_count != job_count) {
        do_logging(LOG_ERR, "FAIL, not all server jobs succeeded!\n");
    }
    do_logging(LOG_ERR, "job_count: %d client_fail_count: %d client_timeout_count: %d server_fail_count: %d client_success_count: %d server_success_count: %d\n", job_count, client_fail_count, client_timeout_count, server_fail_count, client_success_count, server_success_count);

    assert_int_equal(client_fail_count, 0);
    assert_int_equal(client_timeout_count, 0);
    assert_int_equal(server_fail_count, 0);
    assert_int_equal(server_success_count, job_count);
    assert_int_equal(client_success_count, job_count);

    PR_Cleanup();
}


int
main (void)
{
    const struct CMUnitTest tests[] = {
        cmocka_unit_test(ns_stress_test),
    };
    return cmocka_run_group_tests(tests, NULL, NULL);
}

